@@ -13,19 +13,20 @@
 // limitations under the License.
 
 use std::fmt::Debug;
+use std::io::Result;
 use std::sync::Arc;
 
 use async_trait::async_trait;
 
-use crate::error::Result;
 use crate::ops::OpDelete;
 use crate::ops::OpList;
 use crate::ops::OpRead;
 use crate::ops::OpStat;
 use crate::ops::OpWrite;
-use crate::BoxedAsyncReader;
-use crate::BoxedObjectStream;
+use crate::BytesReader;
+use crate::BytesWriter;
 use crate::Metadata;
+use crate::ObjectStreamer;
 
 /// Underlying trait of all backends for implementors.
 ///
@@ -35,19 +36,21 @@ use crate::Metadata;
 /// use [`Operator`][crate::Operator] instead.
 #[async_trait]
 pub trait Accessor: Send + Sync + Debug {
-    /// Read data from the underlying storage into input writer.
-    async fn read(&self, args: &OpRead) -> Result<BoxedAsyncReader> {
+    /// Invoke the `read` operation on the specified path, returns a
+    /// [`BytesReader`][crate::BytesReader] if operate successful.
+    async fn read(&self, args: &OpRead) -> Result<BytesReader> {
         let _ = args;
         unimplemented!()
     }
-    /// Write data from input reader to the underlying storage.
-    async fn write(&self, r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
-        let (_, _) = (r, args);
+    /// Invoke the `write` operation on the specified path, returns a
+    /// [`BytesWriter`][crate::BytesWriter] if operate successful.
+    async fn write(&self, args: &OpWrite) -> Result<BytesWriter> {
+        let _ = args;
         unimplemented!()
     }
     /// Invoke the `stat` operation on the specified path.
     ///
-    /// ## Behavior
+    /// # Behavior
     ///
     /// - `Stat` empty path means stat backend's root path.
     /// - `Stat` a path endswith "/" means stating a dir.
@@ -59,7 +62,7 @@ pub trait Accessor: Send + Sync + Debug {
     }
     /// `Delete` will invoke the `delete` operation.
     ///
-    /// ## Behavior
+    /// # Behavior
     ///
     /// - `Delete` is an idempotent operation, it's safe to call `Delete` on the same path multiple times.
     /// - `Delete` will return `Ok(())` if the path is deleted successfully or not exist.
@@ -68,7 +71,7 @@ pub trait Accessor: Send + Sync + Debug {
         unimplemented!()
     }
 
-    async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
+    async fn list(&self, args: &OpList) -> Result<ObjectStreamer> {
         let _ = args;
         unimplemented!()
     }
@@ -78,11 +81,11 @@ pub trait Accessor: Send + Sync + Debug {
 /// `Accessor` for `Arc<dyn Accessor>`.
 #[async_trait]
 impl<T: Accessor> Accessor for Arc<T> {
-    async fn read(&self, args: &OpRead) -> Result<BoxedAsyncReader> {
+    async fn read(&self, args: &OpRead) -> Result<BytesReader> {
         self.as_ref().read(args).await
     }
-    async fn write(&self, r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
-        self.as_ref().write(r, args).await
+    async fn write(&self, args: &OpWrite) -> Result<BytesWriter> {
+        self.as_ref().write(args).await
     }
     async fn stat(&self, args: &OpStat) -> Result<Metadata> {
         self.as_ref().stat(args).await
@@ -90,7 +93,7 @@ impl<T: Accessor> Accessor for Arc<T> {
     async fn delete(&self, args: &OpDelete) -> Result<()> {
         self.as_ref().delete(args).await
     }
-    async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
+    async fn list(&self, args: &OpList) -> Result<ObjectStreamer> {
         self.as_ref().list(args).await
     }
 }
