@@ -11,6 +11,8 @@
 // WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 // See the License for the specific language governing permissions and
 // limitations under the License.
-pub mod backend;
+mod backend;
 pub use backend::Backend;
 pub use backend::Builder;
+
+mod object_stream;
