@@ -13,6 +13,7 @@
 // limitations under the License.
 
 use std::collections::HashMap;
+use std::io::Result;
 use std::io::SeekFrom;
 use std::path::PathBuf;
 use std::sync::Arc;
@@ -20,10 +21,8 @@ use std::sync::Arc;
 use anyhow::anyhow;
 use async_compat::Compat;
 use async_trait::async_trait;
-use futures::io;
 use futures::AsyncReadExt;
 use futures::AsyncSeekExt;
-use futures::AsyncWriteExt;
 use log::debug;
 use log::error;
 use log::info;
@@ -33,19 +32,20 @@ use tokio::fs;
 
 use super::error::parse_io_error;
 use super::object_stream::Readdir;
-use crate::error::Error;
-use crate::error::Kind;
-use crate::error::Result;
-use crate::object::BoxedObjectStream;
+use crate::error::other;
+use crate::error::BackendError;
+use crate::error::ObjectError;
 use crate::object::Metadata;
 use crate::object::ObjectMode;
+use crate::object::ObjectStreamer;
 use crate::ops::OpDelete;
 use crate::ops::OpList;
 use crate::ops::OpRead;
 use crate::ops::OpStat;
 use crate::ops::OpWrite;
 use crate::Accessor;
-use crate::BoxedAsyncReader;
+use crate::BytesReader;
+use crate::BytesWriter;
 
 #[derive(Default, Debug)]
 pub struct Builder {
@@ -67,11 +67,10 @@ impl Builder {
             None => "/".to_string(),
             Some(v) => {
                 if !v.starts_with('/') {
-                    return Err(Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: HashMap::from([("root".to_string(), v.clone())]),
-                        source: anyhow!("Root must start with /"),
-                    });
+                    return Err(other(BackendError::new(
+                        HashMap::from([("root".to_string(), v.clone())]),
+                        anyhow!("Root must start with /"),
+                    )));
                 }
                 v.to_string()
             }
@@ -128,7 +127,7 @@ impl Backend {
 #[async_trait]
 impl Accessor for Backend {
     #[trace("read")]
-    async fn read(&self, args: &OpRead) -> Result<BoxedAsyncReader> {
+    async fn read(&self, args: &OpRead) -> Result<BytesReader> {
         increment_counter!("opendal_fs_read_requests");
 
         let path = self.get_abs_path(&args.path);
@@ -157,7 +156,7 @@ impl Accessor for Backend {
             })?;
         };
 
-        let r: BoxedAsyncReader = match args.size {
+        let r: BytesReader = match args.size {
             Some(size) => Box::new(f.take(size)),
             None => Box::new(f),
         };
@@ -166,11 +165,11 @@ impl Accessor for Backend {
             "object {} reader created: offset {:?}, size {:?}",
             &path, args.offset, args.size
         );
-        Ok(r)
+        Ok(Box::new(r))
     }
 
     #[trace("write")]
-    async fn write(&self, mut r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
+    async fn write(&self, args: &OpWrite) -> Result<BytesWriter> {
         increment_counter!("opendal_fs_write_requests");
 
         let path = self.get_abs_path(&args.path);
@@ -184,7 +183,13 @@ impl Accessor for Backend {
         //   - Is it better to check the parent dir exists before call mkdir?
         let parent = PathBuf::from(&path)
             .parent()
-            .ok_or_else(|| anyhow!("malformed path: {:?}", &path))?
+            .ok_or_else(|| {
+                other(ObjectError::new(
+                    "write",
+                    &path,
+                    anyhow!("malformed path: {:?}", &path),
+                ))
+            })?
             .to_path_buf();
 
         fs::create_dir_all(&parent).await.map_err(|e| {
@@ -209,27 +214,8 @@ impl Accessor for Backend {
                 e
             })?;
 
-        let mut f = Compat::new(f);
-
-        // TODO: we should respect the input size.
-        let s = io::copy(&mut r, &mut f).await.map_err(|e| {
-            let e = parse_io_error(e, "write", &path);
-            error!("object {} copy: {:?}", &path, e);
-            e
-        })?;
-
-        // `std::fs::File`'s errors detected on closing are ignored by
-        // the implementation of Drop.
-        // So we need to call `flush` to make sure all data have been flushed
-        // to fs successfully.
-        f.flush().await.map_err(|e| {
-            let e = parse_io_error(e, "write", &path);
-            error!("object {} flush: {:?}", &path, e);
-            e
-        })?;
-
         debug!("object {} write finished: size {:?}", &path, args.size);
-        Ok(s as usize)
+        Ok(Box::new(Compat::new(f)))
     }
 
     #[trace("stat")]
@@ -300,7 +286,7 @@ impl Accessor for Backend {
     }
 
     #[trace("list")]
-    async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
+    async fn list(&self, args: &OpList) -> Result<ObjectStreamer> {
         increment_counter!("opendal_fs_list_requests");
 
         let path = self.get_abs_path(&args.path);
