@@ -13,6 +13,10 @@
 // limitations under the License.
 
 use std::collections::HashMap;
+use std::io::Error;
+use std::io::ErrorKind;
+use std::io::Result;
+use std::mem;
 use std::pin::Pin;
 use std::sync::Arc;
 use std::sync::Mutex;
@@ -21,22 +25,22 @@ use std::task::Poll;
 
 use anyhow::anyhow;
 use async_trait::async_trait;
-use bytes::Bytes;
-use futures::io;
-use futures::TryStreamExt;
+use bytes::BufMut;
+use futures::io::Cursor;
+use futures::AsyncWrite;
 use minitrace::trace;
 
-use crate::error::Error;
-use crate::error::Kind;
-use crate::error::Result;
-use crate::object::BoxedObjectStream;
+use crate::error::other;
+use crate::error::ObjectError;
+use crate::object::ObjectStreamer;
 use crate::ops::OpDelete;
 use crate::ops::OpList;
 use crate::ops::OpRead;
 use crate::ops::OpStat;
 use crate::ops::OpWrite;
 use crate::Accessor;
-use crate::BoxedAsyncReader;
+use crate::BytesReader;
+use crate::BytesWriter;
 use crate::Metadata;
 use crate::Object;
 use crate::ObjectMode;
@@ -62,6 +66,10 @@ impl Backend {
 
     // normalize_path removes all internal `//` inside path.
     pub(crate) fn normalize_path(path: &str) -> String {
+        if path.is_empty() || path == "/" {
+            return String::new();
+        }
+
         let has_trailing = path.ends_with('/');
 
         let mut p = path
@@ -81,74 +89,56 @@ impl Backend {
 #[async_trait]
 impl Accessor for Backend {
     #[trace("read")]
-    async fn read(&self, args: &OpRead) -> Result<BoxedAsyncReader> {
+    async fn read(&self, args: &OpRead) -> Result<BytesReader> {
         let path = Backend::normalize_path(&args.path);
 
         let map = self.inner.lock().expect("lock poisoned");
 
-        let data = map.get(&path).ok_or_else(|| Error::Object {
-            kind: Kind::ObjectNotExist,
-            op: "read",
-            path: path.to_string(),
-            source: anyhow!("key not exists in map"),
+        let data = map.get(&path).ok_or_else(|| {
+            Error::new(
+                ErrorKind::NotFound,
+                ObjectError::new("read", &path, anyhow!("key not exists in map")),
+            )
         })?;
 
         let mut data = data.clone();
         if let Some(offset) = args.offset {
             if offset >= data.len() as u64 {
-                return Err(Error::Object {
-                    kind: Kind::Unexpected,
-                    op: "read",
-                    path: path.to_string(),
-                    source: anyhow!("offset out of bound {} >= {}", offset, data.len()),
-                });
+                return Err(other(ObjectError::new(
+                    "read",
+                    &path,
+                    anyhow!("offset out of bound {} >= {}", offset, data.len()),
+                )));
             }
             data = data.slice(offset as usize..data.len());
         };
 
         if let Some(size) = args.size {
             if size > data.len() as u64 {
-                return Err(Error::Object {
-                    kind: Kind::Unexpected,
-                    op: "read",
-                    path: path.to_string(),
-                    source: anyhow!("size out of bound {} > {}", size, data.len()),
-                });
+                return Err(other(ObjectError::new(
+                    "read",
+                    &path,
+                    anyhow!("size out of bound {} > {}", size, data.len()),
+                )));
             }
             data = data.slice(0..size as usize);
         };
 
-        let r: BoxedAsyncReader = Box::new(BytesStream(data).into_async_read());
-        Ok(r)
+        Ok(Box::new(Cursor::new(data)))
     }
+
     #[trace("write")]
-    async fn write(&self, mut r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
+    async fn write(&self, args: &OpWrite) -> Result<BytesWriter> {
         let path = Backend::normalize_path(&args.path);
 
-        let bs = vec![0; args.size as usize];
-        let mut cursor = io::Cursor::new(bs);
-        let n = io::copy(&mut r, &mut cursor)
-            .await
-            .map_err(|e| Error::Object {
-                kind: Kind::Unexpected,
-                op: "write",
-                path: path.clone(),
-                source: anyhow::Error::from(e),
-            })?;
-        if n < args.size {
-            return Err(Error::Object {
-                kind: Kind::Unexpected,
-                op: "write",
-                path: path.clone(),
-                source: anyhow!("write short  {} M {}", n, args.size),
-            });
-        }
-
-        let mut map = self.inner.lock().expect("lock poisoned");
-        map.insert(path.to_string(), Bytes::from(cursor.into_inner()));
-
-        Ok(n as usize)
+        Ok(Box::new(MapWriter {
+            path,
+            size: args.size,
+            map: self.inner.clone(),
+            buf: Default::default(),
+        }))
     }
+
     #[trace("stat")]
     async fn stat(&self, args: &OpStat) -> Result<Metadata> {
         let path = Backend::normalize_path(&args.path);
@@ -165,11 +155,11 @@ impl Accessor for Backend {
 
         let map = self.inner.lock().expect("lock poisoned");
 
-        let data = map.get(&path).ok_or_else(|| Error::Object {
-            kind: Kind::ObjectNotExist,
-            op: "stat",
-            path: path.to_string(),
-            source: anyhow!("key not exists in map"),
+        let data = map.get(&path).ok_or_else(|| {
+            Error::new(
+                ErrorKind::NotFound,
+                ObjectError::new("read", &path, anyhow!("key not exists in map")),
+            )
         })?;
 
         let mut meta = Metadata::default();
@@ -190,7 +180,7 @@ impl Accessor for Backend {
         Ok(())
     }
     #[trace("list")]
-    async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
+    async fn list(&self, args: &OpList) -> Result<ObjectStreamer> {
         let path = Backend::normalize_path(&args.path);
 
         let map = self.inner.lock().expect("lock poisoned");
@@ -209,22 +199,47 @@ impl Accessor for Backend {
     }
 }
 
-struct BytesStream(Bytes);
+struct MapWriter {
+    path: String,
+    size: u64,
+    map: Arc<Mutex<HashMap<String, bytes::Bytes>>>,
 
-impl futures::Stream for BytesStream {
-    type Item = std::result::Result<bytes::Bytes, std::io::Error>;
+    buf: bytes::BytesMut,
+}
 
-    // Always poll the entire stream.
-    fn poll_next(mut self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
-        let size = self.0.len();
-        match self.0.len() {
-            0 => Poll::Ready(None),
-            _ => Poll::Ready(Some(Ok(self.0.split_to(size)))),
-        }
+impl AsyncWrite for MapWriter {
+    fn poll_write(
+        mut self: Pin<&mut Self>,
+        _cx: &mut Context<'_>,
+        buf: &[u8],
+    ) -> Poll<Result<usize>> {
+        let size = buf.len();
+        self.buf.put_slice(buf);
+        Poll::Ready(Ok(size))
     }
 
-    fn size_hint(&self) -> (usize, Option<usize>) {
-        (self.0.len(), Some(self.0.len()))
+    fn poll_flush(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Result<()>> {
+        Poll::Ready(Ok(()))
+    }
+
+    fn poll_close(mut self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Result<()>> {
+        if self.buf.len() != self.size as usize {
+            return Poll::Ready(Err(other(ObjectError::new(
+                "write",
+                &self.path,
+                anyhow!(
+                    "write short, expect {} actual {}",
+                    self.size,
+                    self.buf.len()
+                ),
+            ))));
+        }
+
+        let buf = mem::take(&mut self.buf);
+        let mut map = self.map.lock().expect("lock poisoned");
+        map.insert(self.path.clone(), buf.freeze());
+
+        Poll::Ready(Ok(()))
     }
 }
 
