@@ -14,13 +14,14 @@
 
 use std::cmp::min;
 use std::collections::HashMap;
-use std::fmt::{Debug, Formatter};
+use std::fmt::Debug;
+use std::fmt::Formatter;
+use std::io::Error;
+use std::io::ErrorKind;
+use std::io::Result;
 use std::mem;
-use std::pin::Pin;
 use std::str::FromStr;
 use std::sync::Arc;
-use std::task::Context;
-use std::task::Poll;
 
 use anyhow::anyhow;
 use async_trait::async_trait;
@@ -35,7 +36,6 @@ use hyper::Body;
 use log::debug;
 use log::error;
 use log::info;
-use log::warn;
 use metrics::increment_counter;
 use minitrace::trace;
 use once_cell::sync::Lazy;
@@ -44,21 +44,22 @@ use time::format_description::well_known::Rfc2822;
 use time::OffsetDateTime;
 
 use super::object_stream::S3ObjectStream;
-use crate::credential::Credential;
-use crate::error::Error;
-use crate::error::Kind;
-use crate::error::Result;
-use crate::object::BoxedObjectStream;
+use crate::error::other;
+use crate::error::BackendError;
+use crate::error::ObjectError;
+use crate::io_util::new_http_channel;
+use crate::io_util::HttpBodyWriter;
 use crate::object::Metadata;
-use crate::ops::HeaderRange;
+use crate::object::ObjectStreamer;
+use crate::ops::BytesRange;
 use crate::ops::OpDelete;
 use crate::ops::OpList;
 use crate::ops::OpRead;
 use crate::ops::OpStat;
 use crate::ops::OpWrite;
-use crate::readers::ReaderStream;
 use crate::Accessor;
-use crate::BoxedAsyncReader;
+use crate::BytesReader;
+use crate::BytesWriter;
 use crate::ObjectMode;
 
 /// Allow constructing correct region endpoint if user gives a global endpoint.
@@ -119,9 +120,10 @@ pub struct Builder {
     root: Option<String>,
 
     bucket: String,
-    credential: Option<Credential>,
     endpoint: Option<String>,
     region: Option<String>,
+    access_key_id: Option<String>,
+    secret_access_key: Option<String>,
     server_side_encryption: Option<String>,
     server_side_encryption_aws_kms_key_id: Option<String>,
     server_side_encryption_customer_algorithm: Option<String>,
@@ -135,10 +137,15 @@ impl Debug for Builder {
 
         d.field("root", &self.root)
             .field("bucket", &self.bucket)
-            .field("credential", &self.credential)
             .field("endpoint", &self.endpoint)
             .field("region", &self.region);
 
+        if self.access_key_id.is_some() {
+            d.field("access_key_id", &"<redacted>");
+        }
+        if self.secret_access_key.is_some() {
+            d.field("secret_access_key", &"<redacted>");
+        }
         if self.server_side_encryption.is_some() {
             d.field("server_side_encryption", &"<redacted>");
         }
@@ -180,13 +187,6 @@ impl Builder {
         self
     }
 
-    /// Set credential of this backend.
-    pub fn credential(&mut self, credential: Credential) -> &mut Self {
-        self.credential = Some(credential);
-
-        self
-    }
-
     /// Set endpoint of this backend.
     ///
     /// Endpoint must be full uri, e.g.
@@ -199,11 +199,9 @@ impl Builder {
     /// If user inputs endpoint without scheme like "s3.amazonaws.com", we
     /// will prepend "https://" before it.
     pub fn endpoint(&mut self, endpoint: &str) -> &mut Self {
-        self.endpoint = if endpoint.is_empty() {
-            None
-        } else {
-            Some(endpoint.to_string())
-        };
+        if !endpoint.is_empty() {
+            self.endpoint = Some(endpoint.to_string())
+        }
 
         self
     }
@@ -215,11 +213,33 @@ impl Builder {
     ///
     /// Most of time, region is not need to be set, especially for AWS S3 and minio.
     pub fn region(&mut self, region: &str) -> &mut Self {
-        self.region = if region.is_empty() {
-            None
-        } else {
-            Some(region.to_string())
-        };
+        if !region.is_empty() {
+            self.region = Some(region.to_string())
+        }
+
+        self
+    }
+
+    /// Set access_key_id of this backend.
+    ///
+    /// - If access_key_id is set, we will take user's input first.
+    /// - If not, we will try to load it from environment.
+    pub fn access_key_id(&mut self, v: &str) -> &mut Self {
+        if !v.is_empty() {
+            self.access_key_id = Some(v.to_string())
+        }
+
+        self
+    }
+
+    /// Set secret_access_key of this backend.
+    ///
+    /// - If secret_access_key is set, we will take user's input first.
+    /// - If not, we will try to load it from environment.
+    pub fn secret_access_key(&mut self, v: &str) -> &mut Self {
+        if !v.is_empty() {
+            self.secret_access_key = Some(v.to_string())
+        }
 
         self
     }
@@ -235,11 +255,10 @@ impl Builder {
     /// SSE related options should be set carefully to make them works.
     /// Please use `server_side_encryption_with_*` helpers if even possible.
     pub fn server_side_encryption(&mut self, v: &str) -> &mut Self {
-        self.server_side_encryption = if v.is_empty() {
-            None
-        } else {
-            Some(v.to_string())
-        };
+        if !v.is_empty() {
+            self.server_side_encryption = Some(v.to_string())
+        }
+
         self
     }
 
@@ -261,11 +280,10 @@ impl Builder {
     /// SSE related options should be set carefully to make them works.
     /// Please use `server_side_encryption_with_*` helpers if even possible.
     pub fn server_side_encryption_aws_kms_key_id(&mut self, v: &str) -> &mut Self {
-        self.server_side_encryption_aws_kms_key_id = if v.is_empty() {
-            None
-        } else {
-            Some(v.to_string())
-        };
+        if !v.is_empty() {
+            self.server_side_encryption_aws_kms_key_id = Some(v.to_string())
+        }
+
         self
     }
 
@@ -280,11 +298,10 @@ impl Builder {
     /// SSE related options should be set carefully to make them works.
     /// Please use `server_side_encryption_with_*` helpers if even possible.
     pub fn server_side_encryption_customer_algorithm(&mut self, v: &str) -> &mut Self {
-        self.server_side_encryption_customer_algorithm = if v.is_empty() {
-            None
-        } else {
-            Some(v.to_string())
-        };
+        if !v.is_empty() {
+            self.server_side_encryption_customer_algorithm = Some(v.to_string())
+        }
+
         self
     }
 
@@ -302,11 +319,10 @@ impl Builder {
     /// SSE related options should be set carefully to make them works.
     /// Please use `server_side_encryption_with_*` helpers if even possible.
     pub fn server_side_encryption_customer_key(&mut self, v: &str) -> &mut Self {
-        self.server_side_encryption_customer_key = if v.is_empty() {
-            None
-        } else {
-            Some(v.to_string())
-        };
+        if !v.is_empty() {
+            self.server_side_encryption_customer_key = Some(v.to_string())
+        }
+
         self
     }
 
@@ -323,11 +339,10 @@ impl Builder {
     /// SSE related options should be set carefully to make them works.
     /// Please use `server_side_encryption_with_*` helpers if even possible.
     pub fn server_side_encryption_customer_key_md5(&mut self, v: &str) -> &mut Self {
-        self.server_side_encryption_customer_key_md5 = if v.is_empty() {
-            None
-        } else {
-            Some(v.to_string())
-        };
+        if !v.is_empty() {
+            self.server_side_encryption_customer_key_md5 = Some(v.to_string())
+        }
+
         self
     }
 
@@ -409,11 +424,10 @@ impl Builder {
         let req = hyper::Request::head(format!("{endpoint}/{bucket}"))
             .body(hyper::Body::empty())
             .expect("must be valid request");
-        let res = client.request(req).await.map_err(|e| Error::Backend {
-            kind: Kind::BackendConfigurationInvalid,
-            context: context.clone(),
-            source: anyhow::Error::new(e),
-        })?;
+        let res = client
+            .request(req)
+            .await
+            .map_err(|e| other(BackendError::new(context.clone(), e)))?;
 
         debug!(
             "auto detect region got response: status {:?}, header: {:?}",
@@ -429,11 +443,7 @@ impl Builder {
                     .get("x-amz-bucket-region")
                     .unwrap_or(&HeaderValue::from_static("us-east-1"))
                     .to_str()
-                    .map_err(|e| Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow::Error::new(e),
-                    })?
+                    .map_err(|e| other(BackendError::new(context.clone(), e)))?
                     .to_string();
                 Ok((endpoint.to_string(), region))
             }
@@ -442,25 +452,23 @@ impl Builder {
                 let region = res
                     .headers()
                     .get("x-amz-bucket-region")
-                    .ok_or(Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow!("can't detect region automatically, region is empty"),
+                    .ok_or_else(|| {
+                        other(BackendError::new(
+                            context.clone(),
+                            anyhow!("can't detect region automatically, region is empty"),
+                        ))
                     })?
                     .to_str()
-                    .map_err(|e| Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow::Error::new(e),
-                    })?
+                    .map_err(|e| other(BackendError::new(context.clone(), e)))?
                     .to_string();
-                let template = ENDPOINT_TEMPLATES.get(endpoint).ok_or(Error::Backend {
-                    kind: Kind::BackendConfigurationInvalid,
-                    context: context.clone(),
-                    source: anyhow!(
-                        "can't detect region automatically, no valid endpoint template for {}",
-                        &endpoint
-                    ),
+                let template = ENDPOINT_TEMPLATES.get(endpoint).ok_or_else(|| {
+                    other(BackendError::new(
+                        context.clone(),
+                        anyhow!(
+                            "can't detect region automatically, no valid endpoint template for {}",
+                            &endpoint
+                        ),
+                    ))
                 })?;
 
                 let endpoint = template.replace("{region}", &region);
@@ -469,14 +477,13 @@ impl Builder {
             }
             // Unexpected status code
             code => {
-                return Err(Error::Backend {
-                    kind: Kind::BackendConfigurationInvalid,
-                    context: context.clone(),
-                    source: anyhow!(
+                return Err(other(BackendError::new(
+                    context.clone(),
+                    anyhow!(
                         "can't detect region automatically, unexpected response: status code {}",
                         code
                     ),
-                });
+                )))
             }
         }
     }
@@ -503,11 +510,10 @@ impl Builder {
         // Handle endpoint, region and bucket name.
         let bucket = match self.bucket.is_empty() {
             false => Ok(&self.bucket),
-            true => Err(Error::Backend {
-                kind: Kind::BackendConfigurationInvalid,
-                context: HashMap::from([("bucket".to_string(), "".to_string())]),
-                source: anyhow!("bucket is empty"),
-            }),
+            true => Err(other(BackendError::new(
+                HashMap::from([("bucket".to_string(), "".to_string())]),
+                anyhow!("bucket is empty"),
+            ))),
         }?;
         debug!("backend use bucket {}", &bucket);
 
@@ -527,31 +533,15 @@ impl Builder {
         signer_builder.region(&region);
         signer_builder.allow_anonymous();
 
-        if let Some(cred) = &self.credential {
-            context.insert("credential".to_string(), "*".to_string());
-            match cred {
-                Credential::HMAC {
-                    access_key_id,
-                    secret_access_key,
-                } => {
-                    signer_builder.access_key(access_key_id);
-                    signer_builder.secret_key(secret_access_key);
-                }
-                // We don't need to do anything if user tries to read credential from env.
-                Credential::Plain => {
-                    warn!("backend got empty credential, fallback to read from env.")
-                }
-                _ => {
-                    return Err(Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow!("credential is invalid"),
-                    });
-                }
-            }
+        if let (Some(ak), Some(sk)) = (&self.access_key_id, &self.secret_access_key) {
+            signer_builder.access_key(ak);
+            signer_builder.secret_key(sk);
         }
 
-        let signer = signer_builder.build().await?;
+        let signer = signer_builder
+            .build()
+            .await
+            .map_err(|e| other(BackendError::new(context, e)))?;
 
         info!("backend build finished: {:?}", &self);
         Ok(Arc::new(Backend {
@@ -602,6 +592,11 @@ impl Backend {
 
     // normalize_path removes all internal `//` inside path.
     pub(crate) fn normalize_path(path: &str) -> String {
+        // TODO: we need a unit test.
+        if path.is_empty() || path == "/" {
+            return String::new();
+        }
+
         let has_trailing = path.ends_with('/');
 
         let mut p = path
@@ -707,7 +702,7 @@ impl Backend {
 #[async_trait]
 impl Accessor for Backend {
     #[trace("read")]
-    async fn read(&self, args: &OpRead) -> Result<BoxedAsyncReader> {
+    async fn read(&self, args: &OpRead) -> Result<BytesReader> {
         increment_counter!("opendal_s3_read_requests");
 
         let p = self.get_abs_path(&args.path);
@@ -725,25 +720,39 @@ impl Accessor for Backend {
                     &p, args.offset, args.size
                 );
 
-                Ok(Box::new(ByteStream(resp).into_async_read()))
+                Ok(Box::new(
+                    resp.into_body()
+                        .into_stream()
+                        .map_err(move |e| other(ObjectError::new("read", &p, e)))
+                        .into_async_read(),
+                ))
             }
-            _ => Err(parse_error_response(resp, "read", &p).await),
+            _ => Err(parse_error_response_with_body(resp, "read", &p).await),
         }
     }
+
     #[trace("write")]
-    async fn write(&self, r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
+    async fn write(&self, args: &OpWrite) -> Result<BytesWriter> {
         let p = self.get_abs_path(&args.path);
         debug!("object {} write start: size {}", &p, args.size);
 
-        let resp = self.put_object(&p, r, args.size).await?;
-        match resp.status() {
-            StatusCode::CREATED | StatusCode::OK => {
-                debug!("object {} write finished: size {:?}", &p, args.size);
-                Ok(args.size as usize)
+        let (tx, body) = new_http_channel();
+
+        let req = self.put_object(&p, args.size, body).await;
+
+        let bs = HttpBodyWriter::new(args, tx, self.client.request(req), |op, resp| {
+            match resp.status() {
+                StatusCode::CREATED | StatusCode::OK => {
+                    debug!("object {} write finished: size {:?}", op.path, op.size);
+                    Ok(())
+                }
+                _ => Err(parse_error_response_without_body(resp, "write", &op.path)),
             }
-            _ => Err(parse_error_response(resp, "write", &p).await),
-        }
+        });
+
+        Ok(Box::new(bs))
     }
+
     #[trace("stat")]
     async fn stat(&self, args: &OpStat) -> Result<Metadata> {
         increment_counter!("opendal_s3_stat_requests");
@@ -814,7 +823,7 @@ impl Accessor for Backend {
                 debug!("object {} stat finished", &p);
                 Ok(m)
             }
-            _ => Err(parse_error_response(resp, "stat", &p).await),
+            _ => Err(parse_error_response_with_body(resp, "stat", &p).await),
         }
     }
     #[trace("delete")]
@@ -831,11 +840,11 @@ impl Accessor for Backend {
                 debug!("object {} delete finished", &p);
                 Ok(())
             }
-            _ => Err(parse_error_response(resp, "delete", &p).await),
+            _ => Err(parse_error_response_with_body(resp, "delete", &p).await),
         }
     }
     #[trace("list")]
-    async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
+    async fn list(&self, args: &OpList) -> Result<ObjectStreamer> {
         increment_counter!("opendal_s3_list_requests");
 
         let mut path = self.get_abs_path(&args.path);
@@ -862,7 +871,7 @@ impl Backend {
         if offset.is_some() || size.is_some() {
             req = req.header(
                 http::header::RANGE,
-                HeaderRange::new(offset, size).to_string(),
+                BytesRange::new(offset, size).to_string(),
             );
         }
 
@@ -877,12 +886,7 @@ impl Backend {
 
         self.client.request(req).await.map_err(|e| {
             error!("object {} get_object: {:?}", path, e);
-            Error::Object {
-                kind: Kind::Unexpected,
-                op: "read",
-                path: path.to_string(),
-                source: anyhow::Error::from(e),
-            }
+            other(ObjectError::new("read", path, e))
         })
     }
 
@@ -890,9 +894,9 @@ impl Backend {
     pub(crate) async fn put_object(
         &self,
         path: &str,
-        r: BoxedAsyncReader,
         size: u64,
-    ) -> Result<hyper::Response<hyper::Body>> {
+        body: hyper::Body,
+    ) -> hyper::Request<hyper::Body> {
         let mut req = hyper::Request::put(&format!("{}/{}/{}", self.endpoint, self.bucket, path));
 
         // Set content length.
@@ -902,21 +906,11 @@ impl Backend {
         req = self.insert_sse_headers(req, true);
 
         // Set body
-        let mut req = req
-            .body(hyper::body::Body::wrap_stream(ReaderStream::new(r)))
-            .expect("must be valid request");
+        let mut req = req.body(body).expect("must be valid request");
 
         self.signer.sign(&mut req).await.expect("sign must success");
 
-        self.client.request(req).await.map_err(|e| {
-            error!("object {} put_object: {:?}", path, e);
-            Error::Object {
-                kind: Kind::Unexpected,
-                op: "write",
-                path: path.to_string(),
-                source: anyhow::Error::from(e),
-            }
-        })
+        req
     }
 
     #[trace("head_object")]
@@ -934,12 +928,7 @@ impl Backend {
 
         self.client.request(req).await.map_err(|e| {
             error!("object {} head_object: {:?}", path, e);
-            Error::Object {
-                kind: Kind::Unexpected,
-                op: "stat",
-                path: path.to_string(),
-                source: anyhow::Error::from(e),
-            }
+            other(ObjectError::new("stat", path, e))
         })
     }
 
@@ -954,12 +943,7 @@ impl Backend {
 
         self.client.request(req).await.map_err(|e| {
             error!("object {} delete_object: {:?}", path, e);
-            Error::Object {
-                kind: Kind::Unexpected,
-                op: "delete",
-                path: path.to_string(),
-                source: anyhow::Error::from(e),
-            }
+            other(ObjectError::new("delete", path, e))
         })
     }
 
@@ -985,35 +969,37 @@ impl Backend {
 
         self.client.request(req).await.map_err(|e| {
             error!("object {} list_object: {:?}", path, e);
-            Error::Object {
-                kind: Kind::Unexpected,
-                op: "list",
-                path: path.to_string(),
-                source: anyhow::Error::from(e),
-            }
+            other(ObjectError::new("list", path, e))
         })
     }
 }
 
-struct ByteStream(hyper::Response<hyper::Body>);
+// Read and decode whole error response.
+fn parse_error_response_without_body(resp: Response<Body>, op: &'static str, path: &str) -> Error {
+    let (part, _) = resp.into_parts();
+    let kind = match part.status {
+        StatusCode::NOT_FOUND => ErrorKind::NotFound,
+        StatusCode::FORBIDDEN => ErrorKind::PermissionDenied,
+        _ => ErrorKind::Other,
+    };
 
-impl futures::Stream for ByteStream {
-    type Item = std::io::Result<bytes::Bytes>;
-
-    fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
-        Pin::new(self.0.body_mut())
-            .poll_next(cx)
-            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e.to_string()))
-    }
+    Error::new(
+        kind,
+        ObjectError::new(op, path, anyhow!("response part: {:?}", part)),
+    )
 }
 
 // Read and decode whole error response.
-async fn parse_error_response(resp: Response<Body>, op: &'static str, path: &str) -> Error {
+async fn parse_error_response_with_body(
+    resp: Response<Body>,
+    op: &'static str,
+    path: &str,
+) -> Error {
     let (part, mut body) = resp.into_parts();
     let kind = match part.status {
-        StatusCode::NOT_FOUND => Kind::ObjectNotExist,
-        StatusCode::FORBIDDEN => Kind::ObjectPermissionDenied,
-        _ => Kind::Unexpected,
+        StatusCode::NOT_FOUND => ErrorKind::NotFound,
+        StatusCode::FORBIDDEN => ErrorKind::PermissionDenied,
+        _ => ErrorKind::Other,
     };
 
     // Only read 4KiB from the response to avoid broken services.
@@ -1029,20 +1015,28 @@ async fn parse_error_response(resp: Response<Body>, op: &'static str, path: &str
                     break;
                 }
             }
-            Err(e) => return Error::Unexpected(anyhow!("parse error response parse: {:?}", e)),
+            Err(e) => {
+                return other(ObjectError::new(
+                    op,
+                    path,
+                    anyhow!("parse error response parse: {:?}", e),
+                ))
+            }
         }
     }
 
-    Error::Object {
+    Error::new(
         kind,
-        op,
-        path: path.to_string(),
-        source: anyhow!(
-            "response part: {:?}, body: {:?}",
-            part,
-            String::from_utf8_lossy(&bs)
+        ObjectError::new(
+            op,
+            path,
+            anyhow!(
+                "response part: {:?}, body: {:?}",
+                part,
+                String::from_utf8_lossy(&bs)
+            ),
         ),
-    }
+    )
 }
 
 #[cfg(test)]
